"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const boot_1 = require("@loopback/boot");
const rest_explorer_1 = require("@loopback/rest-explorer");
const repository_1 = require("@loopback/repository");
const rest_1 = require("@loopback/rest");
const service_proxy_1 = require("@loopback/service-proxy");
const path = require("path");
const sequence_1 = require("./sequence");
const joi_1 = require("@hapi/joi");
const _ = require("lodash");
const services_1 = require("./services");
const fs = require("fs");
const yargs = require("yargs");
class GqlApplication extends boot_1.BootMixin(service_proxy_1.ServiceMixin(repository_1.RepositoryMixin(rest_1.RestApplication))) {
    constructor(options = {}) {
        super(options);
        this.argv = null;
        // Parse args and load OAS
        this.run();
        // Set up the custom sequence
        this.sequence(sequence_1.MySequence);
        // Set up default home page
        this.static('/', path.join(__dirname, '../public'));
        // Customize @loopback/rest-explorer configuration here
        if (process.env['ICGQL_ENV'] === 'dev') {
            this.bind(rest_explorer_1.RestExplorerBindings.CONFIG).to({
                path: '/explorer',
                useSelfHostedSpec: false,
            });
            this.component(rest_explorer_1.RestExplorerComponent);
        }
        this.projectRoot = __dirname;
        // Customize @loopback/boot Booter Conventions here
        this.bootOptions = {
            controllers: {
                // Customize ControllerBooter Conventions here
                dirs: ['controllers'],
                extensions: ['.controller.js'],
                nested: true,
            },
        };
    }
    run() {
        this.init()
            .then(() => { })
            .catch(() => { });
    }
    async init() {
        this.parseArgs();
        this.setDefaultArgs();
        this.openApi = new services_1.OpenApiService();
        this.graphQL = new services_1.GraphQLService(this.getSync('opsMap'));
        if (this.argv.validate) {
            await this.openApi.validate(this.argv.oas);
        }
    }
    async loadOAS() {
        let oasData = [];
        let openAPISchema = [];
        oasData = await this.openApi.load(this.argv.oas);
        openAPISchema = await this.openApi.parse(oasData);
        if (this.argv.param) {
            this.openApi.addParameters(openAPISchema, this.argv.param);
        }
        const ops = this.getSync('opsMap');
        const { schema: graphQLSchema, report: graphQLReport, } = await this.graphQL.createSchema(this.argv, openAPISchema, ops);
        if (this.argv.report) {
            console.log('Stats:\n\n');
            console.log(_.omit(graphQLReport, 'warnings'));
            if (graphQLReport.warnings) {
                console.log('There are warnings:\n');
                for (const warn of graphQLReport.warnings) {
                    console.log(`
          Message: ${warn.message}
          Mitigation: ${warn.mitigation}
          `);
                }
            }
        }
        this.bind('oasSchema').to(openAPISchema);
        this.bind('gqlSchema').to(graphQLSchema);
        this.bind('gqlReport').to(graphQLReport);
    }
    setDefaultArgs() {
        const toOpArray = function (type) {
            return function (value) {
                const data = value.split(':');
                return { type, method: data[0], path: data[1] };
            };
        };
        const toObject = (arg) => {
            const ret = {};
            for (const item of arg) {
                const pair = item.split(':');
                ret[pair[0]] = pair[1];
            }
            return ret;
        };
        this.restServer.config.port = this.argv.port;
        this.restServer.config.host = this.argv.host;
        if (!this.argv.oas) {
            const protocol = this.argv.key ? 'https' : 'http';
            this.argv.oas = [
                `${protocol}://${this.argv.host}:${this.argv.port}/api/openapi.json`,
            ];
        }
        if (this.argv.header) {
            this.argv.headerObj = toObject(this.argv.header);
        }
        if (this.argv.queryStr) {
            this.argv.queryObj = toObject(this.argv.queryStr);
        }
        let ops = [];
        if (this.argv.query) {
            ops = [...this.argv.query.map(toOpArray('query'))];
        }
        if (this.argv.mutation) {
            ops = [...ops, ...this.argv.mutation.map(toOpArray('mutation'))];
        }
        this.bind('opsMap').to(ops);
        this.bind('serverKey').to(this.argv.key ? fs.readFileSync(this.argv.key).toString() : null);
        this.bind('serverCert').to(this.argv.cert ? fs.readFileSync(this.argv.cert).toString() : null);
        return this;
    }
    parseArgs() {
        const uriSchema = joi_1.string().uri({ scheme: ['http', 'https'] });
        const toArray = function (arg) {
            return Array.isArray(arg) ? arg : [arg];
        };
        const kvPairTest = (v) => /^[a-z0-9_\-]+:[a-z0-9_\-\/]+$/i.test(v);
        const kvTrippleTest = (v) => /^(header|query|path|cookie):[a-z0-9_\-\/]+:[a-z0-9_\-\/]+$/i.test(v);
        const urlOrFileTest = (v) => {
            const isURL = Boolean(!uriSchema.validate(v).error);
            return isURL || fs.existsSync(v);
        };
        this.argv = yargs
            .option('baseUrl', {
            type: 'string',
            describe: 'The base url',
        })
            .option('oas', {
            type: 'string',
            describe: 'Path or URL of an OAS 2 / 3 YAML or JSON spec. You may supply this flag multiple times.',
        })
            .option('key', {
            type: 'string',
            describe: 'Path to key file. Used by the HTTPS server.',
        })
            .option('cert', {
            type: 'string',
            describe: 'Path to key file. Used by the HTTPS server.',
        })
            .option('host', {
            type: 'string',
            default: '127.0.0.1',
            describe: 'The host address of the server. Should be a fully qualified domain or IP.',
        })
            .option('port', {
            type: 'number',
            default: 3000,
            describe: 'The port number the server should be listening at.',
        })
            .option('limit', {
            type: 'boolean',
            default: true,
            describe: 'Wether or not to add limit to lists',
        })
            .option('fill', {
            type: 'boolean',
            default: true,
            describe: 'Wether or not to fill empty responses',
        })
            .option('strict', {
            type: 'boolean',
            default: false,
            describe: 'Wether or not to run in strict mode',
        })
            .option('errors', {
            type: 'boolean',
            default: true,
            describe: 'Wether or not print debug info and extra verbose op descriptions.',
        })
            .option('header', {
            type: 'string',
            describe: 'A custom HTTP header to be passed to each request generated by GQL. Format: headerName:value. You may supply this flag multiple times.',
        })
            .option('queryStr', {
            type: 'string',
            describe: 'A custom HTTP query string parameter to be passed to each request generated by GQL. Format: name:value. You may supply this flag multiple times',
        })
            .option('query', {
            type: 'string',
            describe: 'Treat this op as query. Argument is the HTTP method - path separated by colon. You may supply this flag multiple times.',
        })
            .option('mutation', {
            type: 'string',
            describe: 'Treat this op as mutation. Argument is the HTTP method - path separated by colon. You may supply this flag multiple times.',
        })
            .option('report', {
            type: 'boolean',
            default: false,
            describe: 'Print report after generating GraphQL schema.',
        })
            .option('validate', {
            type: 'boolean',
            default: false,
            describe: 'Strict OAS validation. If false, only JSON / YAML conformity will be checked.',
        })
            .option('param', {
            type: 'string',
            describe: 'A custom parameter to be added to each OAS operation. Format: type:schema:name. A required param will be added to the OAS parameters object. Values must specified by the user, at runtime. You may supply this flag multiple times.',
        })
            .implies('key', 'cert')
            .implies('cert', 'key')
            .normalize('key')
            .normalize('cert')
            .coerce('header', toArray)
            .coerce('queryStr', toArray)
            .coerce('query', toArray)
            .coerce('mutation', toArray)
            .coerce('oas', toArray)
            .coerce('param', toArray)
            .group(['host', 'port', 'key', 'cert'], 'Server options:')
            .group(['oas'], 'OpenAPI options:')
            .group(['validate'], 'Validation options:')
            .group(['limit', 'fill', 'strict', 'baseUrl'], 'GraphQL options:')
            .group(['queryStr', 'header'], 'Options for generated requests:')
            .group(['query', 'mutation'], 'Transformation options:')
            .group(['report', 'errors'], 'Debug and reporting options:')
            .check((argv) => {
            setTimeout(() => {
                if (!argv.oas.every(urlOrFileTest)) {
                    throw new Error(`OAS is not URL or existing path: "${argv.oas.join(',')}"`);
                }
                else if (argv.key && !fs.existsSync(argv.key)) {
                    throw new Error(`The key file doesn't exist: "${argv.key}"`);
                }
                else if (argv.cert && !fs.existsSync(argv.cert)) {
                    throw new Error(`The certificate file doesn't exist: "${argv.cert}"`);
                }
                else if (argv.baseUrl && uriSchema.validate(argv.baseUrl).error) {
                    throw new Error(`The base URL is invalid: "${argv.baseUrl}"`);
                }
                else if (argv.header && !argv.header.every(kvPairTest)) {
                    throw new Error(`One or more header is not fromatted  correctly: "${argv.header.join(',')}"`);
                }
                else if (argv.queryStr && !argv.queryStr.every(kvPairTest)) {
                    throw new Error(`One or more query param is not fromatted  correctly: "${argv.queryStr.join(',')}"`);
                }
                else if (argv.query && !argv.query.every(kvPairTest)) {
                    throw new Error(`One or more query mappings are not fromatted correctly: "${argv.query.join(',')}"`);
                }
                else if (argv.mutation && !argv.mutation.every(kvPairTest)) {
                    throw new Error(`One or more mutation mappings are not fromatted correctly: "${argv.mutation.join(',')}"`);
                }
                else if (argv.param && !argv.param.every(kvTrippleTest)) {
                    throw new Error(`One or more OAS params are not fromatted correctly: "${argv.param.join(',')}"`);
                }
            });
            return true;
        }, true)
            .locale('pirate')
            .env('ICGQL')
            .completion('autocomplete')
            .usage('npm start -- [options]')
            .example('npm start -- --baseUrl http://127.0.0.1:3000/api', 'Node server mode')
            .example('npm start -- --oas http://some.domain.com:10001/openapi  --header Content-Type:application/json', 'GraphQL layer covering a single REST API')
            .example('npm start -- --oas http://some.domain.com:10001/openapi -o http://some.domain.com:10008/openapi -baseUrl http://some.domain.com --header sessionToken:someToken --header login:someUser --header Content-Type:application/json', 'GraphQL layer covering multiple REST APIs and auth')
            .showHelpOnFail(true)
            .wrap(yargs.terminalWidth())
            .epilogue(`Copyright by i-Cell Mobilsoft ${new Date().getFullYear()}. All rights reserved. See http://github.com/icellmobilsoft/graphqlserver`)
            .help().argv;
        return this;
    }
}
exports.GqlApplication = GqlApplication;
//# sourceMappingURL=application.js.map