"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const application_1 = require("./application");
const express = require("express");
const p_event_1 = require("p-event");
const graphqlHTTP = require('express-graphql');
const bodyParser = require('body-parser');
const helmet = require('helmet');
const https = require("https");
const NoIntrospection = require("graphql-disable-introspection");
// const csurf = require('csurf');
// const session = require('cookie-session');
class ExpressServer {
    constructor(options = {}) {
        var _a;
        this.env = (_a = process.env['ICGQL_ENV'], (_a !== null && _a !== void 0 ? _a : 'prod')).toLowerCase();
        this.app = express();
        this.lbApp = new application_1.GqlApplication(Object.assign(Object.assign({}, options), {
            rest: {
                openApiSpec: {
                    disabled: process.env['ICGQL_NOAPI'] === 'true',
                    endpointMapping: {
                        '/openapi.json': { version: '3.0.0', format: 'json' },
                        '/openapi.yaml': { version: '3.0.0', format: 'yaml' },
                        '/openapi': { version: '3.0.0', format: 'yaml' },
                    },
                },
                apiExplorer: {
                    disabled: this.isProd(),
                },
            },
        }));
    }
    isProd() {
        return this.env === 'prod';
    }
    addMiddlewares() {
        // TODO: Start here, and uncomment if session is needed...
        var _a;
        // const sessionExpiry = new Date(Date.now() + 60 * 60 * 1000);
        // this.app.use(session({
        //   name: 'session',
        //   keys: ['sessionToken'],
        //   cookie: {
        //     secure: true,
        //     httpOnly: true,
        //     domain: 'some.hu',
        //     path: '/',
        //     expires: sessionExpiry
        //   }
        // })
        // );
        this.app.use(helmet({
            noCache: true,
            permittedCrossDomainPolicies: false,
        }));
        const cspConfig = {
            directives: {
                defaultSrc: ["'self'"],
                scriptSrc: !this.isProd()
                    ? ["'unsafe-inline'", "'unsafe-eval'", "'self'"]
                    : ["'none'"],
                styleSrc: !this.isProd() ? ["'unsafe-inline'", "'self'"] : ["'none'"],
                imgSrc: !this.isProd() ? ['data:', "'self'"] : ["'none'"],
                mediaSrc: ["'none'"],
                objectSrc: ["'none'"],
                upgradeInsecureRequests: true,
                workerSrc: ["'none'"],
                fontSrc: ["'none'"],
                frameSrc: ["'none'"],
                frameAncestors: ["'none'"],
                formAction: ["'none'"],
                childSrc: ["'none'"],
                connectSrc: ["'self'"],
                blockAllMixedContent: true,
            },
        };
        if (this.isProd()) {
            cspConfig.directives.requireSriFor = ['script', 'style'];
            process.env['DEBUG'] = (_a = process.env['DEBUG'], (_a !== null && _a !== void 0 ? _a : 'http'));
        }
        this.app.use(helmet.contentSecurityPolicy(cspConfig));
        //this.lbApp.mountExpressRouter('/', csurf());
        this.app.use(helmet.referrerPolicy({ policy: 'no-referrer' }));
        this.app.use(bodyParser.json());
        this.app.use('/api', this.lbApp.requestHandler);
        this.app.disable('x-powered-by');
        this.app.get('/', function (req, res) {
            res.json({ status: 'OK' });
        });
    }
    addGraphQLMiddleware() {
        let gqlConfig = {
            schema: this.lbApp.getSync('gqlSchema'),
        };
        if (!this.isProd()) {
            gqlConfig = Object.assign(Object.assign({}, gqlConfig), {
                graphiql: true,
                pretty: true,
            });
        }
        else {
            gqlConfig.validationRules = [NoIntrospection];
        }
        this.app.use('/graphql', graphqlHTTP(gqlConfig));
    }
    async boot() {
        await this.lbApp.boot();
        this.addMiddlewares();
    }
    async start() {
        await this.lbApp.start();
        const httpsOpts = {
            key: this.lbApp.getSync('serverKey'),
            cert: this.lbApp.getSync('serverCert'),
        };
        if (httpsOpts.key) {
            this.server = https.createServer(httpsOpts, this.app);
            this.server.listen(this.lbApp.restServer.config.port, this.lbApp.restServer.config.host);
        }
        else {
            this.server = this.app.listen(this.lbApp.restServer.config.port, this.lbApp.restServer.config.host);
        }
        await p_event_1.default(this.server, 'listening');
        await this.lbApp.loadOAS();
        this.addGraphQLMiddleware();
        this.app.get('*', function (req, res) {
            res.redirect('/');
        });
    }
    // For testing purposes
    async stop() {
        if (!this.server)
            return;
        await this.lbApp.stop();
        this.server.close();
        await p_event_1.default(this.server, 'close');
        this.server = undefined;
    }
}
exports.ExpressServer = ExpressServer;
//# sourceMappingURL=server.js.map